;****************************************************************************
;                                RAYCAST.ASM
;
;  Assembler-Modul zur Raycasting-Engine.
;  Linkbar mit Turbo-Pascal 7.0.
;  Geschrieben fr Turbo-Assembler 1.01
;
;  Copyright (c) 1998 by Jan Strter.
;
;****************************************************************************

TITLE RAYCAST.ASM

MODEL LARGE,PASCAL

LOCALS @@

DATASEG

;------------------------------------------
; Externe Variablen
;------------------------------------------
extrn VESA_Width: Word
extrn VESA_BufAdr: DWord
extrn VESAx1: Word
extrn VESAy1: Word
extrn WinHeight: DWord
extrn LevelMapOfs: DWord
extrn MapX: DWord
extrn MapY: DWord
extrn PosX: DWord
extrn PosY: DWord

;------------------------------------------
; Konstanten
;------------------------------------------
LevelMapWidth EQU 32
MulFac        EQU 512
MFxWW         EQU MulFac*64
SHRFac        EQU 9

CODESEG
.386P

;****************************************************************************
;****************************************************************************
; Prozedur zum Zeichnen einer texturierten Linie (Vertikal!)
; Verndert keine globalen Variablen.
;****************************************************************************
;****************************************************************************

PUBLIC DrawTexturedLine
PROC DrawtexturedLine FAR WallX: DWord, ScreenX: DWord, Height: DWord, Source: DWord

; Lokale Variablen
     LOCAL Len: DWord, StartY: DWord, ScaleRatio: DWord, SavedFraction: DWord

; Register sichern
     pushad
     push    gs

; Lnge der vertikalen Linie berechnen (Height * 2)
     mov     eax,Height
     shl     eax,1
     mov     Len,eax

; Anfangsadressen laden (ESI: Sprite, EDI: Bildschirmpuffer)
     xor     eax,eax
     mov     gs,ax
     mov     SavedFraction,eax    ; Wute nicht, wohin sonst damit...
     mov     esi,Source
     mov     edi,VESA_BufAdr

; Verhltnis von Quell-/Zielline berechnen (16-Bit Genauigkeit)
     mov     eax,64
     shl     eax,16
     xor     edx,edx
     div     Len
     mov     ScaleRatio,eax

; StartY berechnen (wird ggf. noch korrigiert)
     mov     eax,WinHeight
     shr     eax,1
     sub     eax,Height
     mov     StartY,eax

; Prfen, ob Linie ber dem Bildschirm ragt
     mov     eax,WinHeight
     cmp     Len,eax
     jbe     @@NoClipping

; Ja, also Clipping durchfhren (auch gleichzeitig Quell-Start-Adresse berechnen)
     mov     eax,Len
     mov     ebx,WinHeight
     sub     eax,ebx
     mov     Len,ebx
     mov     StartY,0
     shr     eax,1
     mul     ScaleRatio
     mov     SavedFraction,eax
     and     SavedFraction,65535   ; Die oberen 16 Bit ausmaskieren
     shr     eax,16
     imul    eax,64
     add     esi,eax

; Startadresse am Bildschirm berechnen (unter Verwendung von StartY)
@@NoClipping:
     xor     eax,eax
     mov     ax,VESAy1
     add     eax,StartY
     mov     StartY,eax
     xor     eax,eax
     mov     ax,VESA_Width
     mul     StartY
     xor     ebx,ebx
     mov     bx,VESAx1
     add     eax,ebx
     add     eax,ScreenX
     add     edi,eax

; Startadresse der Textur berechnen
     add     esi,WallX

; Feststellen, ob gezoomt oder verkleinert werden mu
     mov     eax,ScaleRatio
     cmp     eax,65536           ; Bei 16-Bit Nachkomma ist 65536 = 1
     jb      @@Zoom

; Die Quell-Linie wird gestaucht (ScaleRatio >= 1)
; -----------------------------------------------------
; AL:  Farbwert
; AH:  Zhler - uere Schleife (Len - max. 64, daher pat es in ein Byte)
; EBX: Nachkommateil
; ECX: Mindestsprungweite
; EDX: Rest von ScaleRatio Div 65536
; ESI: Textur
; EDI: Bildschirm-Puffer
; EBP: VESA_Width
; -----------------------------------------------------
     mov     eax,ScaleRatio
     mov     ebx,65536           ; Bei 16-Bit Nachkomma ist 65536 = 1
     div     ebx
     shl     eax,6
     mov     ecx,eax             ; Ergebnis nach ECX, EDX wird durch DIV gesetzt
     mov     ebx,Len
     mov     ah,bl
     push    ebp
     xor     ebx,ebx
     mov     bx,VESA_Width
     mov     ebp,ebx
     mov     ebx,65536
@@ReduceLoop:
     mov     al,gs:[esi]
     mov     gs:[edi],al
     add     edi,ebp
     dec     ah
     je      @@RestoreEBP
     sub     ebx,edx
     ja      @@SkipNormal
     add     ebx,65536
     add     esi,ecx
     add     esi,64
     jmp     @@ReduceLoop
@@SkipNormal:
     add     esi,ecx
     jmp     @@ReduceLoop

; Die Quell-Linie wird gezoomt (ScaleRatio < 1)
; -----------------------------------------------------
; EBX: Nachkommaanteil
; ECX: Zhler
; EDX: ScaleRatio
; ESI: Textur
; EDI: Bildschirm-Puffer
; EBP: VESA_Width
; -----------------------------------------------------
@@Zoom:
     mov     ecx,Len
     mov     edx,ScaleRatio
     mov     ebx,SavedFraction
     neg     ebx
     push    ebp
     xor     eax,eax
     mov     ax,VESA_Width
     mov     ebp,eax
@@ZoomExternalLoop:
     add     ebx,65536
     mov     al,gs:[esi]
     add     esi,64
@@ZoomInnerLoop:
     mov     gs:[edi],al
     add     edi,ebp
     dec     ecx
     je      @@RestoreEBP
     sub     ebx,edx
     js      @@ZoomExternalLoop
     jmp     @@ZoomInnerLoop

; Register restaurieren und zurck zur aufrufenden Routine
@@RestoreEBP:
     pop     ebp
@@End:
     pop     gs
     popad
     ret

ENDP



;****************************************************************************
;****************************************************************************
; Prozedur zum Verfolgen des Sehstrahls.
; Verfolgt solange, bis etwas getroffen wird.
; Verndert MapX und MapY.
;****************************************************************************
;****************************************************************************

PUBLIC TraceRay
PROC TraceRay FAR Sin: DWord, Cos: DWord

     LOCAL   AbsCos: DWord, AbsSin: DWord, IncX: DWord, IncY: DWord, IncFac: DWord, AddX: Word, AddY: Word

     pushad

; MapX und MapY auf die aktuelle Position setzen
     mov     eax, PosX
     shr     eax, SHRFac+6
     mov     MapX, eax
     mov     eax, PosY
     shr     eax, SHRFac+6
     mov     MapY, eax

; Vorzeichen von Sinus umkehren
     mov     eax, Sin
     neg     eax
     mov     Sin, eax

; Adresse der Level-Karte laden
     les     si, LevelMapOfs

; Startadresse berechnen
     mov     eax, MapX
     imul    eax, LevelMapWidth
     add     eax, MapY
     shl     eax, 2
     add     si, ax

; Betrge von Sinus und Cosinus bilden
     mov     eax, Cos
     cmp     eax, 0
     jge     @@Proceed1
     neg     eax
@@Proceed1:
     mov     AbsCos, eax
     mov     eax, Sin
     cmp     eax, 0
     jge     @@Proceed2
     neg     eax
@@Proceed2:
     mov     AbsSin, eax

; Prfen, ob Cosinus gro genug ist, wenn nicht, vereinfachte Strahlenverfolgung
     cmp     AbsCos, 5
     jge     @@XLoop
     mov     bx, 4
     mov     ecx, 1
     cmp     Sin, 0
     jge     @@SimpleCosLoop
     neg     bx
     neg     ecx
@@SimpleCosLoop:
     add     si, bx
     add     MapY, ecx
     mov     al, es:[si]
     or      al, al
     je      @@SimpleCosLoop
     jmp     @@WallFound

@@XLoop:

; Prfen, ob Abs(Cos) < Abs(Sin) Oder Abs(Sin) < 5, wenn nicht Y-Schleife verwenden
     mov     eax, AbsSin
     cmp     AbsCos, eax
     jl      @@Proceed0
     cmp     AbsSin, 5
     jl      @@Proceed0
     jmp     @@YLoop
@@Proceed0:

; -------------------------------------------------
;
; Strahlenverfolgung nach Steigungsfkt. - X-Richtung
;
; EBX:  Steigung
; ECX:  X-Koordinate
; DI:   Y-Additionsfaktor
; -------------------------------------------------

; Steigung berechnen
     mov     edx, 0
     mov     eax, MulFac
     imul    Sin
     idiv    Cos
     mov     ebx, eax

; Abhngig vom Vorzeichen von Cos Variablen und Register initialisieren
     cmp     Cos, 0
     jge     @@Proceed3
     mov     IncX, -1
     mov     IncFac, -MFxWW
     mov     AddX, -LevelMapWidth*4
     mov     eax, PosX
     mov     edx, 0
     mov     ecx, MFxWW
     div     ecx
     mov     ecx, edx
     neg     ecx
     jmp     @@Proceed4
@@Proceed3:
     mov     IncX, 1
     mov     IncFac, MFxWW
     mov     AddX, LevelMapWidth*4
     mov     eax, PosX
     mov     edx, 0
     mov     ecx, MFxWW
     div     ecx
     sub     ecx, edx
@@Proceed4:

; Abhging vom Vorzeichen von Sin Variablen und Register initialisieren
     cmp     Sin, 0
     jge     @@Proceed5
     mov     IncY, -1
     mov     di, -4
     jmp     @@Proceed6
@@Proceed5:
     mov     IncY, 1
     mov     di, 4
@@Proceed6:

@@XLoopCore:

; Entfernung zur nchsten Y-Koordinate berechnen (Steigung * X)
     mov     edx, 0
     mov     eax, ebx
     imul    ecx
     shrd    eax, edx, SHRFac
     add     eax, PosY
     sar     eax, SHRFac+6

; Die berechnete Anzahl Kacheln in die richtige Richtung abklappern, wenn Wand gefunden, raus!
@@InnerXLoopCore:
     cmp     eax, MapY
     je      @@EndOfInnerXLoopCore
     add     si, di
     mov     edx, IncY
     add     MapY, edx
     mov     dl, es:[si]
     or      dl, dl
     jne     @@WallFound
     jmp     @@InnerXLoopCore
@@EndOfInnerXLoopCore:

; Einen Schritt in X-Richtung, falls Wand gefunden, raus!
     add     ecx, IncFac
     add     si, AddX
     mov     edx, IncX
     add     MapX, edx
     mov     dl, es:[si]
     or      dl, dl
     jne     @@WallFound
     jmp     @@XLoopCore

@@YLoop:

; -------------------------------------------------
;
; Strahlenverfolgung nach Steigungsfkt. - Y-Richtung
;
; EBX:  Steigung
; ECX:  Y-Koordinate
; DI:   X-Additionsfaktor
; -------------------------------------------------

; Steigung berechnen
     mov     edx, 0
     mov     eax, MulFac
     imul    Cos
     idiv    Sin
     mov     ebx, eax

; Abhngig vom Vorzeichen von Cos Variablen und Register initialisieren
     cmp     Cos, 0
     jge     @@Proceed7
     mov     IncX, -1
     mov     di, -LevelMapWidth*4
     jmp     @@Proceed8
@@Proceed7:
     mov     IncX, 1
     mov     di, LevelMapWidth*4
@@Proceed8:

; Abhging vom Vorzeichen von Sin Variablen und Register initialisieren
     cmp     Sin, 0
     jge     @@Proceed9
     mov     IncY, -1
     mov     AddY, -4
     mov     IncFac, -MFxWW
     mov     eax, PosY
     mov     edx, 0
     mov     ecx, MFxWW
     div     ecx
     mov     ecx, edx
     neg     ecx
     jmp     @@Proceed10
@@Proceed9:
     mov     IncY, 1
     mov     AddY, 4
     mov     eax, PosY
     mov     edx, 0
     mov     ecx, MFxWW
     div     ecx
     sub     ecx, edx
     mov     IncFac, MFxWW
@@Proceed10:

@@YLoopCore:

; Entfernung zur nchsten Y-Koordinate berechnen (Steigung * Y)
     mov     edx, 0
     mov     eax, ebx
     imul    ecx
     shrd    eax, edx, SHRFac
     add     eax, PosX
     sar     eax, SHRFac+6

; Die berechnete Anzahl Kacheln in die richtige Richtung abklappern, wenn Wand gefunden, raus!
@@InnerYLoopCore:
     cmp     eax, MapX
     je      @@EndOfInnerYLoopCore
     add     si, di
     mov     edx, IncX
     add     MapX, edx
     mov     dl, es:[si]
     or      dl, dl
     jne     @@WallFound
     jmp     @@InnerYLoopCore
@@EndOfInnerYLoopCore:

; Einen Schritt in Y-Richtung, falls Wand gefunden, raus!
     add     ecx, IncFac
     add     si, AddY
     mov     edx, IncY
     add     MapY, edx
     mov     dl, es:[si]
     or      dl, dl
     jne     @@WallFound
     jmp     @@YLoopCore

@@WallFound:
     popad
     ret
ENDP



END